@@ -341,8 +341,7 @@ gst_flac_enc_class_init (GstFlacEncClass * klass)
           DEFAULT_SEEKPOINTS,
           G_PARAM_READWRITE | G_PARAM_CONSTRUCT | G_PARAM_STATIC_STRINGS));
 
-  gst_element_class_add_pad_template (gstelement_class,
-      gst_static_pad_template_get (&src_factory));
+  gst_element_class_add_static_pad_template (gstelement_class, &src_factory);
 
   sink_caps = gst_flac_enc_generate_sink_caps ();
   sink_templ = gst_pad_template_new ("sink",
@@ -612,10 +611,12 @@ gst_flac_enc_set_metadata (GstFlacEnc * flacenc, GstAudioInfo * info,
     GstCaps *caps;
     const GstStructure *structure;
     GstTagImageType image_type = GST_TAG_IMAGE_TYPE_NONE;
-    gint i;
+    gint i, width = 0, height = 0, png_icon_count = 0, other_icon_count = 0;
     GstMapInfo map;
 
     for (i = 0; i < n_images + n_preview_images; i++) {
+      gboolean is_preview_image = (i >= n_images);
+
       if (i < n_images) {
         if (!gst_tag_list_get_sample_index (copy, GST_TAG_IMAGE, i, &sample))
           continue;
@@ -636,17 +637,42 @@ gst_flac_enc_set_metadata (GstFlacEnc * flacenc, GstAudioInfo * info,
       flacenc->meta[entries] =
           FLAC__metadata_object_new (FLAC__METADATA_TYPE_PICTURE);
 
+      GST_LOG_OBJECT (flacenc, "image info: %" GST_PTR_FORMAT, structure);
+
       if (structure)
         gst_structure_get (structure, "image-type", GST_TYPE_TAG_IMAGE_TYPE,
             &image_type, NULL);
       else
         image_type = GST_TAG_IMAGE_TYPE_NONE;
 
+      GST_LOG_OBJECT (flacenc, "image caps: %" GST_PTR_FORMAT, caps);
+
+      structure = gst_caps_get_structure (caps, 0);
+      gst_structure_get (structure, "width", G_TYPE_INT, &width,
+          "height", G_TYPE_INT, &height, NULL);
+
       /* Convert to ID3v2 APIC image type */
-      if (image_type == GST_TAG_IMAGE_TYPE_NONE)
-        image_type = (i < n_images) ? 0x00 : 0x01;
-      else
+      if (image_type == GST_TAG_IMAGE_TYPE_NONE) {
+        if (is_preview_image) {
+          /* 1 - 32x32 pixels 'file icon' (PNG only)
+           * 2 - Other file icon
+           * There may only be one each of picture type 1 and 2 in a file. */
+          if (width == 32 && height == 32
+              && gst_structure_has_name (structure, "image/png")
+              && png_icon_count++ == 0) {
+            image_type = 1;
+          } else if (width <= 32 && height <= 32 && other_icon_count++ == 0) {
+            image_type = 2;
+          } else {
+            image_type = 0;     /* Other */
+          }
+        } else {
+          image_type = 0;       /* Other */
+        }
+      } else {
+        /* GStreamer enum is the same but without the two icon types 1+2 */
         image_type = image_type + 2;
+      }
 
       buffer = gst_sample_get_buffer (sample);
       gst_buffer_map (buffer, &map, GST_MAP_READ);
@@ -654,10 +680,13 @@ gst_flac_enc_set_metadata (GstFlacEnc * flacenc, GstAudioInfo * info,
           map.data, map.size, TRUE);
       gst_buffer_unmap (buffer, &map);
 
-      /* FIXME: There's no way to set the picture type in libFLAC */
+      GST_LOG_OBJECT (flacenc, "Setting picture type %d", image_type);
       flacenc->meta[entries]->data.picture.type = image_type;
 
-      structure = gst_caps_get_structure (caps, 0);
+      if (width > 0 && height > 0) {
+        flacenc->meta[entries]->data.picture.width = width;
+        flacenc->meta[entries]->data.picture.height = height;
+      }
 
       FLAC__metadata_object_picture_set_mime_type (flacenc->meta[entries],
           (char *) gst_structure_get_name (structure), TRUE);
@@ -766,9 +795,8 @@ gst_flac_enc_getcaps (GstAudioEncoder * enc, GstCaps * filter)
 
   pad = GST_AUDIO_ENCODER_SINK_PAD (enc);
 
-  if (gst_pad_has_current_caps (pad)) {
-    ret = gst_pad_get_current_caps (pad);
-  } else {
+  ret = gst_pad_get_current_caps (pad);
+  if (ret == NULL) {
     ret = gst_pad_get_pad_template_caps (pad);
   }
 
@@ -813,6 +841,26 @@ done:
   return duration;
 }
 
+static gint64
+gst_flac_enc_get_latency (GstFlacEnc * flacenc)
+{
+  /* The FLAC specification states that the data is processed in blocks,
+   * regardless of the number of channels. Thus, The latency can be calculated
+   * using the blocksize and rate. For example a 1 second block sampled at
+   * 44.1KHz has a blocksize of 44100 */
+
+  /* Get the blocksize */
+  const guint blocksize = FLAC__stream_encoder_get_blocksize (flacenc->encoder);
+
+  /* Get the sample rate in KHz */
+  const guint rate = FLAC__stream_encoder_get_sample_rate (flacenc->encoder);
+  if (!rate)
+    return 0;
+
+  /* Calculate the latecy */
+  return (blocksize * GST_SECOND) / rate;
+}
+
 static gboolean
 gst_flac_enc_set_format (GstAudioEncoder * enc, GstAudioInfo * info)
 {
@@ -859,7 +907,9 @@ gst_flac_enc_set_format (GstAudioEncoder * enc, GstAudioInfo * info)
   if (init_status != FLAC__STREAM_ENCODER_INIT_STATUS_OK)
     goto failed_to_initialize;
 
-  /* no special feedback to base class; should provide all available samples */
+  /* feedback to base class */
+  gst_audio_encoder_set_latency (enc,
+      gst_flac_enc_get_latency (flacenc), gst_flac_enc_get_latency (flacenc));
 
   return TRUE;
 
@@ -1281,9 +1331,8 @@ gst_flac_enc_sink_query (GstAudioEncoder * enc, GstQuery * query)
     case GST_QUERY_ACCEPT_CAPS:{
       GstCaps *acceptable, *caps;
 
-      if (gst_pad_has_current_caps (pad)) {
-        acceptable = gst_pad_get_current_caps (pad);
-      } else {
+      acceptable = gst_pad_get_current_caps (pad);
+      if (acceptable == NULL) {
         acceptable = gst_pad_get_pad_template_caps (pad);
       }
 
@@ -1413,6 +1462,8 @@ gst_flac_enc_set_property (GObject * object, guint prop_id,
     const GValue * value, GParamSpec * pspec)
 {
   GstFlacEnc *this = GST_FLAC_ENC (object);
+  GstAudioEncoder *enc = GST_AUDIO_ENCODER (object);
+  guint64 curr_latency = 0, old_latency = gst_flac_enc_get_latency (this);
 
   GST_OBJECT_LOCK (this);
 
@@ -1480,6 +1531,11 @@ gst_flac_enc_set_property (GObject * object, guint prop_id,
   }
 
   GST_OBJECT_UNLOCK (this);
+
+  /* Update latency if it has changed */
+  curr_latency = gst_flac_enc_get_latency (this);
+  if (old_latency != curr_latency)
+    gst_audio_encoder_set_latency (enc, curr_latency, curr_latency);
 }
 
 static void
